// vim:syntax=ragel


%%{
    machine JSON_mapping;
    alphtype int;
    write data;

    action parse_value {
	state->level++;
	i = _parse_JSON(fpc, pe, state);
	state->level--;

	if (i == NULL) {
	    goto failure;
	}
#ifndef MAPPING_USES_STACK
	else {
	    mapping_insert(m, &(Pike_sp[-2]), &(Pike_sp[-1]));
	    pop_2_elems();
	}
#endif

	c++;
	fexec i;
    }

    action parse_key {
	state->level++;
	i = _parse_JSON_string(fpc, pe, state);
	state->level--;

	if (i == NULL) {
	    goto failure;
	}

	c++;
	fexec i;
    }

    myspace = [ \n\r\t];
    value_start = ["[{\-+.tf] | digit;

    main := '{' . myspace* . (
			start: (
			    '}' -> final |
			    '"' >parse_key . myspace* . ':' -> value
			),
			value: (
			    myspace* . value_start >parse_value . myspace* -> repeat
		        ),
			repeat: (
			    ',' . myspace* -> start |
			    '}' -> final
			)
		       ) %*{ fbreak; };
}%%

p_wchar2 *_parse_JSON_mapping(p_wchar2 *p, p_wchar2 *pe, struct parser_state *state) {
    p_wchar2 *i = p;
#ifndef MAPPING_USES_STACK
    struct mapping *m;
#endif
    int cs;
    int c = 0;

#ifndef MAPPING_USES_STACK
    if (!state->validate) {
	m = debug_allocate_mapping(5);
    }
#endif

    %% write init;
    %% write exec;

    if (cs >= JSON_mapping_first_final) {
	if (!state->validate) {
#ifdef MAPPING_USES_STACK
	    f_aggregate_mapping(c);
#else
	    push_mapping(m);
#endif
	}
	return p;
    }

    push_int((int)p);

failure:
    if (!state->validate) {
#ifdef MAPPING_USES_STACK
	if (c > 0) stack_pop_n_elems_keep_top(c);
#else
	if (c & 1) stack_pop_keep_top(); // remove key
	free_mapping(m);
#endif
#ifdef JUMP
	Pike_error("Error parsing mapping at '%c'.\n", (char)*p);
#endif
    }

    return NULL;
}

